#include "AC33Mx128.h"
#include "aa_types.h"
#include "flash.h"

#define	T_TMR(x)				((x)*20/256)	//t=TMR*256/20MHz(IOSC16)

#define T_ERASE					T_TMR(2500)		//2.5ms
#define T_PGM					T_TMR(2500)		//2.5ms
#define T_DISCHARGE				T_TMR(20)		//20us

//====================================================================
// FMMR (Flash Memory Mode Register)
//
//					0x4000_0104
//
//====================================================================
#define BV_IDLE					(0x0001UL<<24)
//#define WDTCON_WDBG			(0x0001<<15)

//====================================================================
// FMCR (Flash Memory Control Register)
//
//					0x4000_0108
//
//====================================================================
#define BV_HRESPD				(0x0001UL<<31)
//
#define BV_PCLK2				(0x0001UL<<27)
#define BV_CLK4					(0x0001UL<<25)
#define BV_CLK3					(0x0001UL<<24)
//
#define BV_TIMER				(0x0001UL<<20)
//
#define BV_EVER					(0x0001UL<<14)
#define BV_PVER					(0x0001UL<<13)
//
#define BV_AE					(0x0001UL<<8)
//
#define BV_PMODE				(0x0001UL<<5)
#define BV_WE					(0x0001UL<<4)
//
#define BV_PBLD					(0x0001UL<<3)
#define BV_PGM					(0x0001UL<<2)
#define BV_ERS					(0x0001UL<<1)
#define BV_PBR					(0x0001UL<<0)



//#define	flash_cmd	(*_flash_cmd)
//int		img_flash_cmd(int dsel,unsigned long cmd, unsigned long addr, FWORD *pdata);
//int		(*_flash_cmd)(int dsel,unsigned long cmd, unsigned long addr, FWORD *pdata);

uint32_t	_fmcr_bak;

//#define	FLASH_COPY_SIZE			(512)
//static	unsigned char	_flash_cmd_ram[FLASH_COPY_SIZE];	//execution buffer
/*
INLINE	void	flash_delay(void)
{
	int	i;
	
	for(i=0;i<10;i++) {
		;//__asm {};
	}
}

void	flash_protect(int prt)
{
	;
}
*/

//-------------------------------------------------------
INLINE	void	flash_entry(void)
{
	FM->MR = 0x81;
	FM->MR = 0x28;
	_fmcr_bak = FM->CR;
	FM->CR = 0x82000000;	  // 0x01000000; 2-wait , 0x02000000; 3-wait

	FM->MR = 0x5A;
	FM->MR = 0xA5;
	FM->DRTY=0;
}

INLINE	void	flash_exit(void)
{
	FM->MR = 0x81;
	FM->MR = 0x28;
	FM->CR = _fmcr_bak;
	FM->MR = 0;
}

//-------------------------------------------------------
int		flash_wait_timer(void);
int		flash_discharge(void);

int		flash_cmd(int dsel,unsigned long cmd, unsigned long addr, FWORD *pdata)
{
	int	i;
	int	result=0;
 	
	flash_entry();
	
	FM->CR = 0;
	FM->CR|= BV_PMODE;  //program mode en
	
	if(flash_wait_timer()) {
		result=-1;
	}
	else {
		//----------------------------------
		if(cmd&BV_PGM) {
			FM->CR|= BV_PBR;  //page buffer reset en
			FM->CR&=~BV_PBR;  //page buffer reset dis
				
			FM->CR|= BV_PBLD;  //page buffer load : en 
			for(i=0;i<SECTOR_SIZE;i+=sizeof(FWORD)) {
				FM->DR = *pdata++;
			}
			FM->CR&=~BV_PBLD;   //page buffer load : dis
		}
		//----------------------------------
		FM->AR = (addr>>2);  // address load 

		FM->CR|= cmd;         // PGM or ERS en
		FM->CR|= BV_WE;     // WE en
		
		if(flash_wait_timer()) {  // wait
			result = -1;
		}

		FM->CR&=~(BV_WE|BV_ERS|BV_PGM);  // PGM, ERS and WE dis
		flash_discharge();
		FM->CR&=~(BV_AE);  // if AE ON, clear AE bit
		flash_discharge();
		
		FM->CR&=~BV_PMODE; //program mode dis
	}
	flash_exit();
	
	return result;
}

INLINE	int		flash_wait_timer(void)
{
	int	i;
	int	t;
	int	error=0;
	
	t=FM->MR;
	for(i=0x10000000;i>0;i--) {
		t=FM->MR;
		if((t & BV_IDLE))
			break;
	}
	if(i==0)
		error|=1;

	return error;
}
INLINE	int		flash_discharge(void)
{
	FM->TMR = T_DISCHARGE;
	FM->CR|= BV_TIMER;
	
	flash_wait_timer();
	
	FM->CR&=~BV_TIMER;
	return 0;
}

//-------------------------------------------------------

int	flash_erase_macro(int dsel)
{
	int	result;
	FM->TMR = T_ERASE;
	result = flash_cmd(dsel,BV_ERS|BV_AE, 0, 0);
	return result;
}

int	flash_erase_sector(int dsel, unsigned long addr)
{
	int	result;
	FM->TMR = T_ERASE;
	result = flash_cmd(dsel,BV_ERS, addr, 0);
	return result;
}

int	flash_write_sector(int dsel, unsigned long addr,FWORD *pdata)
{
	int	result;
	FM->TMR = T_PGM;
	result = flash_cmd(dsel,BV_PGM, addr, pdata);
	return result;
}

/*
int		flash_init(void)
{
	int				i;
	unsigned char	*ptr;
	
	ptr = (unsigned char*)(((unsigned long)&img_flash_cmd) &~1);
	for(i=0;i<FLASH_COPY_SIZE;i++) {
		_flash_cmd_ram[i] = ptr[i];
	}
	_flash_cmd = (void*)((unsigned long)&_flash_cmd_ram | 1);
	
	return 0;
}
*/
